import json
import time
import typing
from openai import AsyncOpenAI
from openai.types.chat import ChatCompletionToolParam
from .base import DEFAULT_MODEL_CONFIGURATION, Conversation, ModelConfiguration
from pydantic import BaseModel


class Tool(BaseModel):
    name: str
    description: str
    parameters: type[BaseModel]


class ToolCall(BaseModel):
    id: str
    name: str
    arguments: dict


class ToolCallCompletion(BaseModel):
    conversation: Conversation
    start_time: float
    end_time: float
    tool_calls: list[ToolCall]


def tool_to_openai_tool(tool: Tool) -> ChatCompletionToolParam:
    model_json_schema = tool.parameters.model_json_schema()
    model_json_schema["additionalProperties"] = False

    return typing.cast(
        ChatCompletionToolParam,
        {
            "type": "function",
            "function": {
                "name": tool.name,
                "description": tool.description,
                "parameters": model_json_schema,
            },
            "strict": True,
        },
    )


async def generate_tool_call_completion_for_conversation(
    conversation: Conversation,
    tools: Tool | list[Tool],
    model_configuration: ModelConfiguration | None = None,
) -> ToolCallCompletion:
    """Generate a tool call completion from a conversation.

    This function sends the conversation to the OpenAI API with the provided tools,
    and returns the tool calls that the model generates.

    Args:
        conversation: The conversation to generate a tool call completion for.
        tools: A Tool object or a list of Tool objects that define the available tools for the model.
        model_configuration: Optional configuration for the model. If None, uses the default configuration.

    Returns:
        A ToolCallCompletion object containing the tool calls generated by the model.
    """
    if model_configuration is None:
        model_configuration = DEFAULT_MODEL_CONFIGURATION

    if isinstance(tools, Tool):
        tools = [tools]

    client = AsyncOpenAI(
        api_key=model_configuration.api_key,
        base_url=model_configuration.base_url,
    )

    openai_tools = [tool_to_openai_tool(tool) for tool in tools]

    start_time = time.time()
    completion = await client.chat.completions.create(
        model=model_configuration.model,
        messages=conversation.openai_messages,
        temperature=model_configuration.temperature,
        logprobs=model_configuration.log_probs,
        top_logprobs=model_configuration.top_log_probs,
        max_tokens=model_configuration.max_tokens,
        presence_penalty=model_configuration.presence_penalty,
        frequency_penalty=model_configuration.frequency_penalty,
        top_p=model_configuration.top_p,
        seed=model_configuration.seed,
        tools=openai_tools,
    )
    end_time = time.time()

    openai_tool_calls = completion.choices[0].message.tool_calls
    if openai_tool_calls is None:
        raise ValueError("No tool calls found in the completion.")

    tool_calls = [
        ToolCall(
            id=tool_call.id,
            name=tool_call.function.name,
            arguments=json.loads(tool_call.function.arguments),
        )
        for tool_call in openai_tool_calls
    ]

    return ToolCallCompletion(
        conversation=conversation,
        start_time=start_time,
        end_time=end_time,
        tool_calls=tool_calls,
    )


async def generate_tool_call_completion(
    user_prompt: str,
    tools: Tool | list[Tool],
    system_prompt: str | None = None,
    model_configuration: ModelConfiguration | None = None,
) -> ToolCallCompletion:
    """
    Generate a tool call completion for a user prompt.

    This function creates a conversation from the provided prompts and generates
    tool calls based on the available tools.

    Args:
        user_prompt: The user prompt to generate a tool call for.
        tools: A Tool object or a list of Tool objects that define the available tools for the model.
        system_prompt: Optional system prompt to use for the conversation.
        model_configuration: Optional configuration for the model. If None, uses the default configuration.

    Returns:
        A ToolCallCompletion object containing the generated tool calls.
    """
    if model_configuration is None:
        model_configuration = DEFAULT_MODEL_CONFIGURATION

    conversation = Conversation.from_prompts(user_prompt, system_prompt=system_prompt)

    return await generate_tool_call_completion_for_conversation(
        conversation,
        tools,
        model_configuration=model_configuration,
    )



from .base import (
    Message,
    TokenLogProb,
    Conversation,
    TextCompletion,
    StructuredCompletion,
    ModelConfiguration,
)

from .completion import (
    generate_text_completion_for_conversation,
    generate_text_completion,
    generate_text_completions_for_conversations,
    generate_text_completions,
)

from .structured_completion import (
    generate_structured_completion_for_conversation,
    generate_structured_completion,
    generate_structured_completions_for_conversations,
    generate_structured_completions,
)

from .tool_call import (
    Tool,
    ToolCall,
    generate_tool_call_completion,
)

__all__ = [
    # From base
    "Message",
    "TokenLogProb",
    "Conversation",
    "TextCompletion",
    "StructuredCompletion",
    "ModelConfiguration",
    # From completion
    "generate_text_completion_for_conversation",
    "generate_text_completion",
    "generate_text_completions_for_conversations",
    "generate_text_completions",
    # From structured_completion
    "generate_structured_completion_for_conversation",
    "generate_structured_completion",
    "generate_structured_completions_for_conversations",
    "generate_structured_completions",
    # From tool_call
    "Tool",
    "ToolCall",
    "generate_tool_call_completion",
]



import math
from typing import Generic, Literal, TypeVar, cast
from openai.types.chat import ChatCompletionMessageParam
from openai.types.chat.chat_completion import Choice
from pydantic import BaseModel, Field

T = TypeVar("T")


def get_first_element(list: list[T]) -> T | None:
    if len(list) == 0:
        return None
    return list[0]


def get_last_element(list: list[T]) -> T | None:
    if len(list) == 0:
        return None
    return list[-1]


class Message(BaseModel):
    role: Literal["system", "user", "assistant"]
    content: str

    @property
    def openai_message(self) -> ChatCompletionMessageParam:
        return cast(
            ChatCompletionMessageParam, {"role": self.role, "content": self.content}
        )


class TokenLogProb(BaseModel):
    token: str
    log_prob: float

    @property
    def prob(self) -> float:
        return math.exp(self.log_prob)

    def __repr__(self) -> str:
        return f"TokenLogProb(token={self.token!r}, prob={round(self.prob, 2)})"


class Conversation(BaseModel):
    messages: list[Message] = Field(default_factory=list)

    def add_message(self, message: Message):
        last_message = get_last_element(self.messages)

        if last_message is None:
            if message.role == "assistant":
                raise ValueError("The first message must be a system or user message")

            self.messages.append(message)
            return

        if last_message.role == "system" and message.role != "user":
            raise ValueError("System message must be followed by a user message")

        if last_message.role == "assistant" and message.role != "user":
            raise ValueError("Assistant message must be followed by a user message")

        if last_message.role == "user" and message.role != "assistant":
            raise ValueError("User message must be followed by an assistant message")

        self.messages.append(message)

    def to_pretty_string(
        self,
        system_color: str = "\033[1;36m",
        user_color: str = "\033[1;32m",
        assistant_color: str = "\033[1;35m",
    ) -> str:
        pretty_lines = []

        for message in self.messages:
            if message.role == "system":
                color_code = system_color
            elif message.role == "user":
                color_code = user_color
            elif message.role == "assistant":
                color_code = assistant_color

            # Reset color code
            reset_code = "\033[0m"

            pretty_lines.append(f"{color_code}{message.role.capitalize()}{reset_code}")

            separator_length = len(message.role) + 2  # +2 for some extra space
            pretty_lines.append("-" * separator_length)

            pretty_lines.append(f"{message.content}\n")

        return "\n".join(pretty_lines).strip()

    def to_markdown(self) -> str:
        markdown_str = ""
        for message in self.messages:
            markdown_str += f"## {message.role.capitalize()} \n"
            markdown_str += f"{message.content}\n\n"
        return markdown_str.strip()

    @property
    def openai_messages(self) -> list[ChatCompletionMessageParam]:
        return [message.openai_message for message in self.messages]

    @staticmethod
    def from_prompts(
        user_prompt: str,
        assistant_prompt: str | None = None,
        system_prompt: str | None = None,
    ) -> "Conversation":
        conversation = Conversation()
        if system_prompt is not None:
            conversation.add_message(Message(role="system", content=system_prompt))
        conversation.add_message(Message(role="user", content=user_prompt))
        if assistant_prompt is not None:
            conversation.add_message(
                Message(role="assistant", content=assistant_prompt)
            )
        return conversation


def format_token_log_probs(
    token_log_probs: list[TokenLogProb], show_probabilities: bool = False
) -> str:
    """
    Returns a pretty string representation of the token log probabilities.
    Tokens are colored from dark red (low probability) to dark green (high probability).

    :param token_log_probs: List of TokenLogProb objects to format
    :param show_probabilities: Whether to show the probability value after each token.
    """
    result = []
    for token_log_prob in token_log_probs:
        if token_log_prob.prob < 0.25:
            color_code = "\033[1;31m"  # Dark red
        elif token_log_prob.prob < 0.5:
            color_code = "\033[1;33m"  # Yellow
        elif token_log_prob.prob < 0.75:
            color_code = "\033[1;32m"  # Light green
        else:
            color_code = "\033[1;92m"  # Dark green

        # Reset color code
        reset_code = "\033[0m"

        if show_probabilities:
            result.append(
                f"{color_code}{token_log_prob.token}[{round(token_log_prob.prob, 2)}]{reset_code}"
            )
        else:
            result.append(f"{color_code}{token_log_prob.token}{reset_code}")

    return "".join(result)


class TextCompletion(BaseModel):
    conversation: Conversation
    model: str
    content: str
    start_time: float
    end_time: float

    """
    A list containing the most likely tokens and their log probabilities for each token position in the message.
    """
    full_token_log_probs: list[list[TokenLogProb]] | None = None

    @property
    def duration(self) -> float:
        """The duration of the generation in seconds."""
        return self.end_time - self.start_time

    @property
    def token_log_probs(self) -> list[TokenLogProb] | None:
        if self.full_token_log_probs is None:
            return None

        return [
            token_log_probs_position[0]
            for token_log_probs_position in self.full_token_log_probs
        ]

    def to_pretty_log_probs_string(self, show_probabilities: bool = False) -> str:
        """
        Returns a pretty string representation of the token log probabilities.
        Tokens are colored from dark red (low probability) to dark green (high probability).

        :param show_probabilities: Whether to show the probability value after each token.
        """
        if self.token_log_probs is None:
            return "No token log probabilities available."

        return format_token_log_probs(self.token_log_probs, show_probabilities)


class StructuredCompletion(BaseModel, Generic[T]):
    conversation: Conversation
    model: str
    content: T
    start_time: float
    end_time: float
    full_token_log_probs: list[list[TokenLogProb]] | None = None

    @property
    def duration(self) -> float:
        """The duration of the generation in seconds."""
        return self.end_time - self.start_time

    @property
    def token_log_probs(self) -> list[TokenLogProb] | None:
        if self.full_token_log_probs is None:
            return None

        return [
            token_log_probs_position[0]
            for token_log_probs_position in self.full_token_log_probs
        ]

    def to_pretty_log_probs_string(self, show_probabilities: bool = False) -> str:
        if self.token_log_probs is None:
            return "No token log probabilities available."

        return format_token_log_probs(self.token_log_probs, show_probabilities)


def parse_logprobs(first_choice: Choice) -> list[list[TokenLogProb]] | None:
    token_log_probs = None
    if first_choice.logprobs is not None and first_choice.logprobs.content is not None:
        token_log_probs = [
            [
                TokenLogProb(
                    token=token_log_prob.token, log_prob=token_log_prob.logprob
                )
                for token_log_prob in log_probs_content.top_logprobs
            ]
            for log_probs_content in first_choice.logprobs.content
        ]
    return token_log_probs


class ModelConfiguration(BaseModel):
    model: str = "gpt-4o"
    temperature: float = 1.0
    log_probs: bool = False
    top_log_probs: int | None = None
    max_tokens: int | None = None
    presence_penalty: float | None = None
    frequency_penalty: float | None = None
    top_p: float | None = None
    seed: int | None = None
    api_key: str | None = None
    base_url: str | None = None


DEFAULT_MODEL_CONFIGURATION = ModelConfiguration()



import asyncio
import time
from typing import Type
from openai import AsyncOpenAI
from tqdm import tqdm

from .base import (
    DEFAULT_MODEL_CONFIGURATION,
    T,
    Conversation,
    ModelConfiguration,
    StructuredCompletion,
    get_first_element,
    parse_logprobs,
)


async def generate_structured_completion_for_conversation(
    conversation: Conversation,
    response_model: Type[T],
    model_configuration: ModelConfiguration | None = None,
) -> StructuredCompletion[T]:
    """
    Generate a structured completion for a conversation.

    :param conversation: The conversation to generate a completion for.
    :param response_model: The model to parse the response into.
    :param model_configuration: The model configuration to use for the completion.
    :return: A StructuredCompletion object.
    """
    if model_configuration is None:
        model_configuration = DEFAULT_MODEL_CONFIGURATION

    client = AsyncOpenAI(
        api_key=model_configuration.api_key,
        base_url=model_configuration.base_url,
    )

    start_time = time.time()
    completion = await client.beta.chat.completions.parse(
        model=model_configuration.model,
        messages=conversation.openai_messages,
        response_format=response_model,
        temperature=model_configuration.temperature,
        logprobs=model_configuration.log_probs,
        top_logprobs=model_configuration.top_log_probs,
        max_tokens=model_configuration.max_tokens,
        presence_penalty=model_configuration.presence_penalty,
        frequency_penalty=model_configuration.frequency_penalty,
        top_p=model_configuration.top_p,
        seed=model_configuration.seed,
    )
    end_time = time.time()

    first_choice = get_first_element(completion.choices)
    if first_choice is None:
        raise ValueError("No choices returned from the completion.")

    message_content = first_choice.message.parsed

    if message_content is None:
        raise ValueError("No message content returned from the completion.")

    full_token_log_probs = parse_logprobs(first_choice)

    return StructuredCompletion(
        conversation=conversation,
        model=model_configuration.model,
        content=message_content,
        start_time=start_time,
        end_time=end_time,
        full_token_log_probs=full_token_log_probs,
    )


async def generate_structured_completion(
    user_prompt: str,
    response_model: Type[T],
    system_prompt: str | None = None,
    model_configuration: ModelConfiguration | None = None,
) -> StructuredCompletion[T]:
    """
    Generate a structured completion for a user prompt.

    :param user_prompt: The user prompt to generate a completion for.
    :param response_model: The model to parse the response into.
    :param system_prompt: The system prompt to use for the completion.
    :param model_configuration: The model configuration to use for the completion.
    :return: A StructuredCompletion object.
    """
    if model_configuration is None:
        model_configuration = DEFAULT_MODEL_CONFIGURATION

    conversation = Conversation.from_prompts(user_prompt, system_prompt=system_prompt)

    return await generate_structured_completion_for_conversation(
        conversation,
        response_model=response_model,
        model_configuration=model_configuration,
    )


async def generate_structured_completions_for_conversations(
    conversations: list[Conversation],
    response_model: Type[T],
    n_parallel: int = 5,
    model_configuration: ModelConfiguration | None = None,
    show_progress: bool = True,
) -> list[StructuredCompletion[T]]:
    """
    Generate structured completions for a list of conversations with support for parallel generation.

    :param conversations: The list of conversations to generate completions for.
    :param response_model: The model to parse the responses into.
    :param n_parallel: The number of completions to generate in parallel.
    :param model_configuration: The model configuration to use for the completions.
    :param show_progress: Whether to show a progress bar.
    :return: A list of StructuredCompletion objects.
    """
    completions = []

    progress_bar = None

    if show_progress:
        progress_bar = tqdm(total=len(conversations))

    for i in range(0, len(conversations), n_parallel):
        conversations_batch = conversations[i : i + n_parallel]

        tasks = [
            asyncio.create_task(
                generate_structured_completion_for_conversation(
                    conversation,
                    response_model=response_model,
                    model_configuration=model_configuration,
                )
            )
            for conversation in conversations_batch
        ]

        completions_batch = await asyncio.gather(*tasks)
        completions.extend(completions_batch)

        if show_progress and progress_bar is not None:
            progress_bar.update(len(completions_batch))

    if show_progress and progress_bar is not None:
        progress_bar.close()

    return completions


async def generate_structured_completions(
    user_prompts: list[str],
    response_model: Type[T],
    system_prompt: str | None = None,
    n_parallel: int = 5,
    model_configuration: ModelConfiguration | None = None,
    show_progress: bool = True,
) -> list[StructuredCompletion[T]]:
    """
    Generate structured completions for a list of user prompts with support for parallel generation.

    :param user_prompts: The list of user prompts to generate completions for.
    :param response_model: The model to parse the responses into.
    :param system_prompt: The system prompt to use for the completions.
    :param n_parallel: The number of completions to generate in parallel.
    :param model_configuration: The model configuration to use for the completions.
    :param show_progress: Whether to show a progress bar.
    :return: A list of StructuredCompletion objects.
    """
    conversations = [
        Conversation.from_prompts(user_prompt, system_prompt=system_prompt)
        for user_prompt in user_prompts
    ]

    return await generate_structured_completions_for_conversations(
        conversations,
        response_model=response_model,
        n_parallel=n_parallel,
        model_configuration=model_configuration,
        show_progress=show_progress,
    )



import pytest

from limin import (
    Conversation,
    Message,
    TextCompletion,
    TokenLogProb,
    StructuredCompletion,
)


def test_message_openai_message():
    message = Message(role="user", content="Test message")
    openai_message = message.openai_message

    assert isinstance(openai_message, dict)
    assert openai_message["role"] == "user"
    assert openai_message["content"] == "Test message"

    expected = {"role": "user", "content": "Test message"}
    assert openai_message == expected


def test_token_log_prob_prob():
    token_log_prob = TokenLogProb(token="test", log_prob=0.0)
    assert token_log_prob.prob == 1.0


def test_conversation_add_message():
    conversation = Conversation()

    system_message = Message(role="system", content="System prompt")
    conversation.add_message(system_message)
    assert len(conversation.messages) == 1
    assert conversation.messages[0].role == "system"

    user_message = Message(role="user", content="Hello")
    conversation.add_message(user_message)
    assert len(conversation.messages) == 2
    assert conversation.messages[1].role == "user"

    assistant_message = Message(role="assistant", content="Hi there")
    conversation.add_message(assistant_message)
    assert len(conversation.messages) == 3
    assert conversation.messages[2].role == "assistant"

    user_message2 = Message(role="user", content="How are you?")
    conversation.add_message(user_message2)
    assert len(conversation.messages) == 4
    assert conversation.messages[3].role == "user"


def test_conversation_add_message_invalid_first_message():
    conversation = Conversation()
    assistant_message = Message(role="assistant", content="Invalid first message")
    with pytest.raises(ValueError):
        conversation.add_message(assistant_message)


def test_conversation_add_message_invalid_message_sequence_user_user():
    conversation = Conversation()
    conversation.add_message(Message(role="user", content="Hello"))
    with pytest.raises(ValueError):
        conversation.add_message(Message(role="user", content="Hi there"))


def test_conversation_add_message_invalid_message_sequence_assistant_assistant():
    conversation = Conversation()
    conversation.add_message(Message(role="user", content="Hello"))
    conversation.add_message(Message(role="assistant", content="Hi there"))
    with pytest.raises(ValueError):
        conversation.add_message(Message(role="assistant", content="Hi there"))


def test_conversation_to_pretty_string():
    conversation = Conversation()

    system_message = Message(role="system", content="System prompt")
    conversation.add_message(system_message)

    user_message = Message(role="user", content="Hello")
    conversation.add_message(user_message)

    assistant_message = Message(role="assistant", content="Hi there")
    conversation.add_message(assistant_message)

    pretty_string = conversation.to_pretty_string()

    assert (
        pretty_string
        == """\033[1;36mSystem\033[0m
--------
System prompt

\033[1;32mUser\033[0m
------
Hello

\033[1;35mAssistant\033[0m
-----------
Hi there"""
    )


def test_conversation_to_markdown():
    conversation = Conversation()

    system_message = Message(role="system", content="System prompt")
    conversation.add_message(system_message)

    user_message = Message(role="user", content="Hello")
    conversation.add_message(user_message)

    assistant_message = Message(role="assistant", content="Hi there")
    conversation.add_message(assistant_message)

    markdown_string = conversation.to_markdown()

    assert (
        markdown_string
        == """## System 
System prompt

## User 
Hello

## Assistant 
Hi there"""
    )


def test_conversation_openai_messages():
    conversation = Conversation()

    system_message = Message(role="system", content="System prompt")
    conversation.add_message(system_message)

    user_message = Message(role="user", content="Hello")
    conversation.add_message(user_message)

    assistant_message = Message(role="assistant", content="Hi there")
    conversation.add_message(assistant_message)

    openai_messages = conversation.openai_messages

    assert len(openai_messages) == 3
    assert openai_messages[0] == {"role": "system", "content": "System prompt"}
    assert openai_messages[1] == {"role": "user", "content": "Hello"}
    assert openai_messages[2] == {"role": "assistant", "content": "Hi there"}


def test_conversation_from_prompts_user_prompt():
    user_prompt = "Hello"
    conversation = Conversation.from_prompts(user_prompt)

    assert len(conversation.messages) == 1
    assert conversation.messages[0].role == "user"
    assert conversation.messages[0].content == "Hello"


def test_conversation_from_prompts_user_assistant_prompt():
    user_prompt = "Hello"
    assistant_prompt = "Hi there"
    conversation = Conversation.from_prompts(user_prompt, assistant_prompt)

    assert len(conversation.messages) == 2
    assert conversation.messages[0].role == "user"
    assert conversation.messages[0].content == "Hello"
    assert conversation.messages[1].role == "assistant"
    assert conversation.messages[1].content == "Hi there"


def test_conversation_from_prompts_user_assistant_system_prompt():
    user_prompt = "Hello"
    assistant_prompt = "Hi there"
    system_prompt = "System prompt"
    conversation = Conversation.from_prompts(
        user_prompt, assistant_prompt, system_prompt
    )

    assert len(conversation.messages) == 3
    assert conversation.messages[0].role == "system"
    assert conversation.messages[0].content == "System prompt"
    assert conversation.messages[1].role == "user"
    assert conversation.messages[1].content == "Hello"
    assert conversation.messages[2].role == "assistant"
    assert conversation.messages[2].content == "Hi there"


def create_text_completion():
    conversation = Conversation.from_prompts("Hello", "Hi there")

    return TextCompletion(
        conversation=conversation,
        model="gpt-4o",
        content="Test content",
        start_time=100.0,
        end_time=105.5,
        full_token_log_probs=[
            [
                TokenLogProb(token="Test", log_prob=-0.1),
                TokenLogProb(token="Example", log_prob=-0.2),
            ],
            [
                TokenLogProb(token="content", log_prob=-0.2),
                TokenLogProb(token="word", log_prob=-0.3),
            ],
        ],
    )


def create_structured_completion():
    conversation = Conversation.from_prompts("Hello", "Hi there")

    return StructuredCompletion(
        conversation=conversation,
        model="gpt-4o",
        content={"key": "value"},
        start_time=100.0,
        end_time=107.0,
        full_token_log_probs=[
            [
                TokenLogProb(token="{", log_prob=-0.1),
                TokenLogProb(token="[", log_prob=-0.3),
            ],
            [
                TokenLogProb(token="key", log_prob=-0.2),
                TokenLogProb(token="name", log_prob=-0.4),
            ],
        ],
    )


def test_text_completion_duration():
    text_completion = create_text_completion()

    assert text_completion.duration == 5.5


def test_text_completion_token_log_probs():
    text_completion = create_text_completion()

    assert len(text_completion.token_log_probs) == 2
    assert text_completion.token_log_probs[0].token == "Test"
    assert text_completion.token_log_probs[0].log_prob == -0.1
    assert text_completion.token_log_probs[1].token == "content"
    assert text_completion.token_log_probs[1].log_prob == -0.2


def test_text_completion_to_pretty_log_probs_string():
    text_completion = create_text_completion()

    pretty_log_probs_string = text_completion.to_pretty_log_probs_string()

    assert pretty_log_probs_string == "\x1b[1;92mTest\x1b[0m\x1b[1;92mcontent\x1b[0m"


def test_text_completion_to_pretty_log_probs_string_with_probabilities():
    text_completion = create_text_completion()

    pretty_log_probs_string = text_completion.to_pretty_log_probs_string(
        show_probabilities=True
    )

    assert (
        pretty_log_probs_string
        == "\x1b[1;92mTest[0.9]\x1b[0m\x1b[1;92mcontent[0.82]\x1b[0m"
    )


def test_structured_completion_duration():
    structured_completion = create_structured_completion()

    assert structured_completion.duration == 7.0


def test_structured_completion_token_log_probs():
    structured_completion = create_structured_completion()

    assert len(structured_completion.token_log_probs) == 2
    assert structured_completion.token_log_probs[0].token == "{"
    assert structured_completion.token_log_probs[0].log_prob == -0.1
    assert structured_completion.token_log_probs[1].token == "key"
    assert structured_completion.token_log_probs[1].log_prob == -0.2


def test_structured_completion_to_pretty_log_probs_string():
    structured_completion = create_structured_completion()

    pretty_log_probs_string = structured_completion.to_pretty_log_probs_string()

    assert pretty_log_probs_string == "\x1b[1;92m{\x1b[0m\x1b[1;92mkey\x1b[0m"


def test_structured_completion_to_pretty_log_probs_string_with_probabilities():
    structured_completion = create_structured_completion()

    pretty_log_probs_string = structured_completion.to_pretty_log_probs_string(
        show_probabilities=True
    )

    assert (
        pretty_log_probs_string == "\x1b[1;92m{[0.9]\x1b[0m\x1b[1;92mkey[0.82]\x1b[0m"
    )
